/**
 * Contentful Delivery API Client. Contains methods which allow access to the
 * different kinds of entities present in Contentful (Entries, Assets, etc).
 */
import { AxiosInstance } from 'contentful-sdk-core';
import { GetGlobalOptions } from './create-global-options';
import { Asset, AssetCollection, AssetFields, AssetKey, AssetQueries, ContentType, ContentTypeCollection, EntriesQueries, EntryCollection, EntryWithoutLinkResolution, EntryCollectionWithoutLinkResolution, LocaleCollection, LocaleCode, EntryWithAllLocalesAndWithoutLinkResolution, EntryCollectionWithAllLocalesAndWithoutLinkResolution, EntryWithLinkResolutionAndWithUnresolvableLinks, EntryCollectionWithLinkResolutionAndWithUnresolvableLinks, EntryWithAllLocalesAndWithLinkResolutionAndWithUnresolvableLinks, EntryCollectionWithAllLocalesAndWithLinkResolutionAndWithUnresolvableLinks, EntryWithLinkResolutionAndWithoutUnresolvableLinks, EntryCollectionWithLinkResolutionAndWithoutUnresolvableLinks, EntryWithAllLocalesAndWithLinkResolutionAndWithoutUnresolvableLinks, EntryCollectionWithAllLocalesAndWithLinkResolutionAndWithoutUnresolvableLinks, Space, SyncCollection, Tag, TagCollection } from './types';
import { EntryQueries } from './types/query/query';
import { FieldsType } from './types/query/util';
import { ChainOptions } from './utils/client-helpers';
export interface ClientWithLinkResolutionAndWithUnresolvableLinks extends BaseClient {
    getEntry<Fields extends FieldsType>(id: string, query?: EntryQueries): Promise<EntryWithLinkResolutionAndWithUnresolvableLinks<Fields>>;
    getEntries<Fields extends FieldsType>(query?: EntriesQueries<Fields>): Promise<EntryCollectionWithLinkResolutionAndWithUnresolvableLinks<Fields>>;
}
export interface ClientWithoutLinkResolution extends BaseClient {
    getEntry<Fields extends FieldsType>(id: string, query?: EntryQueries): Promise<EntryWithoutLinkResolution<Fields>>;
    getEntries<Fields extends FieldsType>(query?: EntriesQueries<Fields>): Promise<EntryCollectionWithoutLinkResolution<Fields>>;
}
export interface ClientWithAllLocalesAndWithLinkResolutionAndWithUnresolvableLinks extends Omit<BaseClient, 'getEntries' | 'getEntry'> {
    getEntry<Fields extends FieldsType = FieldsType, Locales extends LocaleCode = any>(id: string, query?: EntryQueries & {
        locale?: never;
    }): Promise<EntryWithAllLocalesAndWithLinkResolutionAndWithUnresolvableLinks<Fields, Locales>>;
    getEntries<Fields extends FieldsType, Locales extends LocaleCode = any>(query?: EntriesQueries<Fields> & {
        locale?: never;
    }): Promise<EntryCollectionWithAllLocalesAndWithLinkResolutionAndWithUnresolvableLinks<Fields, Locales>>;
}
export interface ClientWithAllLocalesAndWithoutLinkResolution extends Omit<BaseClient, 'getEntries' | 'getEntry'> {
    getEntry<Fields extends FieldsType, Locales extends LocaleCode = any>(id: string, query?: EntryQueries & {
        locale?: never;
    }): Promise<EntryWithAllLocalesAndWithoutLinkResolution<Fields, Locales>>;
    getEntries<Fields extends FieldsType, Locales extends LocaleCode = any>(query?: EntriesQueries<Fields> & {
        locale?: never;
    }): Promise<EntryCollectionWithAllLocalesAndWithoutLinkResolution<Fields, Locales>>;
}
export interface ClientWithLinkResolutionAndWithoutUnresolvableLinks extends Omit<BaseClient, 'getEntries' | 'getEntry'> {
    getEntry<Fields extends FieldsType>(id: string, query?: EntryQueries): Promise<EntryWithLinkResolutionAndWithoutUnresolvableLinks<Fields>>;
    getEntries<Fields extends FieldsType>(query?: EntriesQueries<Fields>): Promise<EntryCollectionWithLinkResolutionAndWithoutUnresolvableLinks<Fields>>;
}
export interface ClientWithAllLocalesAndWithLinkResolutionAndWithoutUnresolvableLinks extends Omit<BaseClient, 'getEntries' | 'getEntry'> {
    getEntry<Fields extends FieldsType, Locales extends LocaleCode = any>(id: string, query?: EntryQueries & {
        locale?: never;
    }): Promise<EntryWithAllLocalesAndWithLinkResolutionAndWithoutUnresolvableLinks<Fields, Locales>>;
    getEntries<Fields extends FieldsType, Locales extends LocaleCode = any>(query?: EntriesQueries<Fields> & {
        locale?: never;
    }): Promise<EntryCollectionWithAllLocalesAndWithLinkResolutionAndWithoutUnresolvableLinks<Fields, Locales>>;
}
export type DefaultClient = ClientWithLinkResolutionAndWithUnresolvableLinks;
export interface BaseClient {
    version: string;
    /**
     * Gets an Asset
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const asset = await client.getAsset('<asset_id>')
     * console.log(asset)
     * ```
     */
    getAsset(id: string, query?: {
        locale?: string;
    }): Promise<Asset>;
    /**
     * Gets a collection of Assets
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getAssets()
     * console.log(response.items)
     * ```
     */
    getAssets(query?: AssetQueries<AssetFields>): Promise<AssetCollection>;
    /**
     * Gets a Content Type
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const contentType = await client.getContentType('<content_type_id>')
     * console.log(contentType)
     * ```
     */
    getContentType(id: string): Promise<ContentType>;
    /**
     * Gets a collection of Content Types
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getContentTypes()
     * console.log(response.items)
     * ```
     */
    getContentTypes(query?: {
        query?: string;
    }): Promise<ContentTypeCollection>;
    /**
     * Gets a collection of Entries
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getEntries()
     * console.log(response.items)
     * ```
     */
    getEntries<Fields extends FieldsType>(query?: EntriesQueries<Fields>): Promise<EntryCollectionWithLinkResolutionAndWithUnresolvableLinks<Fields>>;
    /**
     * Gets an Entry
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const entry = await client.getEntry('<entry_id>')
     * console.log(entry)
     * ```
     */
    getEntry<Fields extends FieldsType>(id: string, query?: EntryQueries): Promise<EntryWithLinkResolutionAndWithUnresolvableLinks<Fields>>;
    /**
     * Gets the Space which the client is currently configured to use
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     * // returns the space object with the above <space-id>
     * const space = await client.getSpace()
     * console.log(space)
     * ```
     */
    getSpace(): Promise<Space>;
    /**
     * Gets a collection of Locales
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getLocales()
     * console.log(response.items)
     * ```
     */
    getLocales(): Promise<LocaleCollection>;
    /**
     * Parse raw json data into collection of entry objects. Link resolution depends on global configuration.
     * @category API
     * @example
     * ```javascript
     * let data = {items: [
     *    {
     *    sys: {type: 'Entry', locale: 'en-US'},
     *    fields: {
     *      animal: {sys: {type: 'Link', linkType: 'Animal', id: 'oink'}},
     *      anotheranimal: {sys: {type: 'Link', linkType: 'Animal', id: 'middle-parrot'}}
     *    }
     *  }
     * ],
     * includes: {
     *  Animal: [
     *    {
     *      sys: {type: 'Animal', id: 'oink', locale: 'en-US'},
     *      fields: {
     *        name: 'Pig',
     *        friend: {sys: {type: 'Link', linkType: 'Animal', id: 'groundhog'}}
     *      }
     *    }
     *   ]
     *  }
     * }
     * console.log( data.items[0].fields.foo ); // undefined
     * let parsedData = client.parseEntries(data);
     * console.log( parsedData.items[0].fields.foo ); // foo
     * ```
     */
    parseEntries<T>(raw: any): EntryCollection<T>;
    /**
     * Synchronizes either all the content or only new content since last sync
     * See <a href="https://www.contentful.com/developers/docs/concepts/sync/">Synchronization</a> for more information.
     * <strong> Important note: </strong> The the sync api endpoint does not support include or link resolution.
     * However contentful.js is doing link resolution client side if you only make an initial sync.
     * For the delta sync (using nextSyncToken) it is not possible since the sdk wont have access to all the data to make such an operation.
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.sync({
     *   initial: true
     * })
     * console.log({
     *   entries: response.entries,
     *   assets: response.assets,
     *   nextSyncToken: response.nextSyncToken
     * })
     * ```
     */
    sync(query: any): Promise<SyncCollection>;
    /**
     * Gets a Tag
     * @category API
     * @example
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const tag = await client.getTag('<asset_id>')
     * console.log(tag)
     */
    getTag(id: string): Promise<Tag>;
    /**
     * Gets a collection of Tags
     * @category API
     * @example
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getTags()
     * console.log(response.items)
     */
    getTags(query?: any): Promise<TagCollection>;
    /**
     * Creates an asset key for signing asset URLs (Embargoed Assets)
     * @category API
     * @example
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const assetKey = await client.getAssetKey(<UNIX timestamp>)
     * console.log(assetKey)
     */
    createAssetKey(expiresAt: number): Promise<AssetKey>;
}
export interface CreateContentfulApiParams {
    http: AxiosInstance;
    getGlobalOptions: GetGlobalOptions;
}
export type ContentfulClientApi = {
    withAllLocales: {
        withoutLinkResolution: ClientWithAllLocalesAndWithoutLinkResolution;
        withoutUnresolvableLinks: ClientWithAllLocalesAndWithLinkResolutionAndWithoutUnresolvableLinks;
    } & ClientWithAllLocalesAndWithLinkResolutionAndWithUnresolvableLinks;
    withoutLinkResolution: {
        withAllLocales: ClientWithAllLocalesAndWithoutLinkResolution;
    } & ClientWithoutLinkResolution;
    withoutUnresolvableLinks: {
        withAllLocales: ClientWithAllLocalesAndWithLinkResolutionAndWithoutUnresolvableLinks;
    } & ClientWithLinkResolutionAndWithoutUnresolvableLinks;
} & DefaultClient;
export default function createContentfulApi<OptionType extends ChainOptions>({ http, getGlobalOptions }: CreateContentfulApiParams, options?: OptionType): DefaultClient;
